@@ -3,11 +3,6 @@
 --
 -- Code for ziggurats.
 --
--- Important notes:
--- ----------------
--- Functions that are attached to Lua markers' onclimb properties
--- cannot be closures, because Lua markers must be saved and closure
--- upvalues cannot (yet) be saved.
 ------------------------------------------------------------------------------
 
 function zig()
@@ -40,11 +35,6 @@ function initialise_ziggurat(z, portal)
   z.colour = ziggurat_wall_colour()
 end
 
-function callback.ziggurat_initialiser(portal)
-  dgn.persist.ziggurat = { }
-  initialise_ziggurat(dgn.persist.ziggurat, portal)
-end
-
 -- Common setup for ziggurat levels.
 function ziggurat_level(e)
   e.tags("allow_dup")
@@ -66,6 +56,10 @@ function ziggurat_awaken_all(mons)
 end
 
 function ziggurat_build_level(e)
+  if you.depth() == 1 then
+    dgn.persist.ziggurat = { }
+    initialise_ziggurat(dgn.persist.ziggurat, portal)
+  end
   local builder = zig().builder
 
   -- Deeper levels can have all monsters awake.
@@ -73,12 +67,6 @@ function ziggurat_build_level(e)
   local generate_awake = you.depth() > 4 + crawl.random2(21)
   zig().monster_hook = generate_awake and global_function("ziggurat_awaken_all")
 
-  -- Deeper levels may block controlled teleports.
-  -- Does never happen at depths 1-6; does always happen at depths 25-27.
-  if you.depth() > 6 + crawl.random2(19) then
-    dgn.change_level_flags("no_tele_control")
-  end
-
   if builder then
     return ziggurat_builder_map[builder](e)
   end
@@ -200,126 +188,280 @@ local function mset_if(condition, ...)
   mset(unpack(util.map(util.curry(spec_if, condition), { ... })))
 end
 
-mset(with_props("place:Lair:$ w:165 / dire elephant w:12 / " ..
-                "catoblepas w:12 / hellephant w:6 / spriggan druid w:1 / " ..
-                "guardian serpent w:1 / deep troll shaman w:1 / " ..
-                "raiju w:1 / hell beast w:1", { weight = 5 }),
-     with_props("place:Shoals:$ w:125 / merfolk aquamancer / water nymph w:5 / " ..
-                "merfolk impaler w:5 / merfolk javelineer", { weight = 5 }),
-     "place:Spider:$ w:115 / ghost moth w:15 / red wasp / " ..
-                "orb spider",
-     "place:Crypt:$ w:260 / curse skull w:5 / profane servitor w:5 / " ..
-                "bone dragon / ancient lich / revenant",
-     "place:Abyss:$ w:1990 / corrupter",
-     with_props("place:Slime:$", { jelly_protect = true }),
-     with_props("place:Coc:$ w:460 / Ice Fiend / " ..
-                 "blizzard demon w:30", { weight = 5 }),
-     with_props("place:Geh:$ w:460 / Brimstone Fiend / " ..
-                 "balrug w:30", { weight = 5 }),
-     with_props("place:Dis:$ w:460 / Hell Sentinel / " ..
-                 "dancing weapon / iron dragon w:20", { weight = 5 }),
-     with_props("place:Tar:$ w:460 / Shadow Fiend / " ..
-                 "curse toe / shadow demon w:20", { weight = 5 }),
-     with_props("daeva / angel / cherub / pearl dragon / " ..
-                "ophan / apis", { weight = 2 }),
-     with_props("hill giant / cyclops / stone giant / fire giant / " ..
-                "frost giant / ettin / titan", { weight = 2 }),
-     with_props("fire elemental / fire drake / hell hound / efreet / " ..
-                "fire dragon / fire giant / orb of fire", { weight = 2 }),
-     with_props("ice beast / freezing wraith / ice dragon / " ..
-                "frost giant / ice devil / ice fiend / simulacrum / " ..
-                "white draconian knight / blizzard demon", { weight = 2 }),
-     with_props("insubstantial wisp / air elemental / titan / raiju / " ..
-                "storm dragon / electric golem / spriggan air mage / " ..
-                "shock serpent", { weight = 2 }),
-     with_props("spectral thing / shadow wraith / eidolon w:4 / shadow dragon / " ..
-                "deep elf death mage w:6 / death knight w:4 / " ..
-                "revenant w:4 / profane servitor w:6 / soul eater / " ..
-                "shadow fiend / black sun", { weight = 2 }),
-     with_props("swamp drake / fire drake / wind drake w:2 / death drake / " ..
-                "wyvern w:5 / hydra w:5 / steam dragon / mottled dragon / " ..
-                "swamp dragon / fire dragon / ice dragon / storm dragon / " ..
-                "iron dragon / shadow dragon / quicksilver dragon / " ..
-                "golden dragon", { weight = 2 }),
-     with_props("centaur / yaktaur / centaur warrior / yaktaur captain / " ..
-                "cyclops / stone giant / faun w:1 / satyr w:2 / " ..
-                "thorn hunter w:2 / naga sharpshooter / " ..
-                "merfolk javelineer / deep elf master archer", { weight = 2 }),
-     with_props("wizard / necromancer / ogre mage w:5 / orc sorcerer w:5 / " ..
-                "naga mage / naga ritualist w:5 / salamander mystic w:5 / " ..
-                "greater naga / tengu conjurer / tengu reaver / " ..
-                "spriggan air mage w:5 / merfolk aquamancer w:5 / " ..
-                "deep elf conjurer / deep elf annihilator / " ..
-                "deep elf sorcerer / draconian scorcher w:5 / " ..
-                "draconian knight w:5 / draconian annihilator w:5 / " ..
-                "lich w:3 / ancient lich w:2 / blood saint", { weight = 2 }))
+-- Monster sets, in order:
+-- Lair, Snake, Swamp, Shoals, Spider, Slime,
+-- Orc, Elf, Vaults, Crypt, Tomb,
+-- Abyss, Gehenna, Cocytus, Dis, Tartarus,
+-- Fire, Ice, Air, Earth, Negative Energy, Holy, Chaos
+-- Giants, Dragons, Draconians, Archers, Conjurers
+-- Pan, Lair Roulette
+-- By using spec_fn to wrap monster-spec functions, monster weights
+-- are adjusted per-set, sometimes scaling by depth and always by zig completion.
 
--- spec_fn can be used to wrap a function that returns a monster spec.
--- This is useful if you want to adjust monster weights in the spec
--- wrt to depth in the ziggurat.
-mset(spec_fn(function ()
-               local d = 290 - 10 * you.depth()
-               local e = math.max(0, you.depth() - 20)
-               return "place:Orc:$ w:" .. d .. " / orc warlord / " ..
-                 "orc high priest band / orc sorcerer w:5 / stone giant / " ..
-                 "iron troll w:5 / moth of wrath w:" .. e
-             end))
+mset(with_props(spec_fn(function ()
+  local d = math.max(24, math.floor(150 - you.depth() / 4 - you.zigs_completed() * 4))
+  local e = 15 + you.zigs_completed()
+  local f = 5 + you.zigs_completed() * 3
+  local g = 10 + you.zigs_completed() * 4
+  return "place:Lair:$ w:" .. d .. " / dire elephant w:" .. e .. " / " ..
+         "hydra w:" .. e .. " / torpor snail w:" .. e - 10 .. " / " ..
+         "catoblepas w:" .. f .. " / spriggan druid w:" .. f .. " / " ..
+         "hellephant w:" .. g .. " / caustic shrike w:" .. g
+end), { weight = 5 }))
 
 mset(spec_fn(function ()
-               local d = 300 - 10 * you.depth()
-               return "place:Elf:$ w:" .. d .. " / deep elf high priest / " ..
-                 "deep elf blademaster / deep elf master archer / " ..
-                 "deep elf annihilator / deep elf demonologist"
-             end))
+  local d = math.max(6, you.depth() * 2 + you.zigs_completed() - 2)
+  local e = math.max(1, you.depth() + you.zigs_completed() - 18)
+  return "place:Snake:$ w:125 / naga sharpshooter w:5 / guardian serpent w:5 / " ..
+         "salamander tyrant w:" .. d .. " / nagaraja w:" .. d .. " / " ..
+         "quicksilver dragon w:" .. e
+end))
 
 mset(spec_fn(function ()
-               local d = math.max(3, you.depth() - 1)
-               local e = math.max(1, you.depth() - 20)
-               return "place:Snake:$ w:65 / guardian serpent w:5 / " ..
-                 "greater naga w:" .. d .. " / quicksilver dragon w:" .. e
-             end))
+  local d = math.max(10, math.min(120 - you.zigs_completed() * 4, 280 - you.depth() * 10))
+  local e = math.floor(10 + you.zigs_completed() / 3)
+  local f = 5 + you.zigs_completed() * 3
+  local g = math.max(1, you.depth() + you.zigs_completed() - 9)
+  return "place:Swamp:$ w:" .. d .. " / hydra w:" .. e .. " / " ..
+         "swamp dragon w:" .. e .. "  / tentacled monstrosity w:" .. f .. " / " ..
+         "shambling mangrove w:" .. f .. " / green death w:" .. g + 3 .. " / " ..
+         "death drake w:" .. g .. " / golden dragon w:" .. g
+end))
+
+mset(with_props(spec_fn(function ()
+  local d = math.max(24, math.floor(120 - you.depth() / 4 - you.zigs_completed() * 4))
+  local e = 10 + you.zigs_completed() * 4
+  return "place:Shoals:$ band w:" .. d .. " / merfolk impaler w:5 / " ..
+         "merfolk javelineer / water nymph w:" .. e .. " / " ..
+         "merfolk aquamancer w:" .. e
+end), { weight = 5 }))
 
 mset(spec_fn(function ()
-               local d = math.max(120, 280 - 10 * you.depth())
-               local e = math.max(1, you.depth() - 9)
-               return "place:Swamp:$ w:" .. d .. " / hydra / " ..
-                 "swamp dragon / green death w:6 / death drake w:1 / " ..
-                 "golden dragon w:1 / tentacled monstrosity w:" .. e
-             end))
+  local d = 10 + you.zigs_completed() * 4
+  local e = 30 + you.zigs_completed() * 5
+  local f = 5 + you.zigs_completed()
+  return "place:Spider:$ w:250 / torpor snail w:" .. d .. " / " ..
+         "emperor scorpion w:" .. d .. " / entropy weaver w:" .. d + 10 .. " / " ..
+         "orb spider w:" .. d + 10 .. " / ghost moth w:" .. e .. " / " ..
+         "moth of wrath w:" .. f
+end))
 
 mset(spec_fn(function ()
-               local d = math.max(1, you.depth() - 11)
-               return "place:Vaults:$ 9 w:30 / place:Vaults:$ w:60 / " ..
-                 "titan w:" .. d .. " / golden dragon w:" .. d ..
-                 " / ancient lich w:" .. d
-             end))
+  local d = 5 + 5 * you.zigs_completed()
+  return "place:Slime:$ w:1500 / floating eye w:" .. d .. " / " ..
+         "azure jelly w:" .. d .. " / quicksilver ooze w:" .. d .. " / " ..
+         "rockslime w:" .. d * 3 .. " / acid blob w:" .. d * 14
+end))
 
 mset(spec_fn(function ()
-               local d = you.depth() + 5
-               return "place:Tomb:$ w:200 / greater mummy w:" .. d
-             end))
+  local d = math.max(2, 290 - 10 * you.depth() - you.zigs_completed())
+  local e = 10 + you.zigs_completed()
+  local f = math.max(0, math.floor(you.depth() + you.zigs_completed() / 2 - 20))
+  return "place:Orc:$ w:" .. d .. " / orc warlord w:" .. e .. " / " ..
+         "orc high priest w:" .. e .. " / orc sorcerer w:5 / " ..
+         "stone giant w:5 / iron troll w:5 / " ..
+         "moth of wrath w:" .. f .. " / juggernaut w:" .. e - 10
+end))
 
 mset(spec_fn(function ()
-               local d = math.max(2, math.floor((32 - you.depth()) / 5))
-               local e = math.min(8, math.floor((you.depth()) / 5) + 4)
-               local f = math.max(1, you.depth() - 5)
-               return "chaos spawn w:" .. d .. " / ugly thing w:" .. d ..
-                 " / very ugly thing w:5 / apocalypse crab w:5 / " ..
-                 "shapeshifter hd:16 w:" ..e .. " / glowing shapeshifter w:" .. e ..
-                 " / killer klown w:8 / chaos champion w:2 / pandemonium lord w:" .. f
-             end))
+  local d = math.max(2, 300 - 10 * you.depth() - you.zigs_completed())
+  local e = math.max(2, math.floor(10 - you.zigs_completed() / 3))
+  return "place:Elf:$ w:" .. d .. " / deep elf high priest / " ..
+         "deep elf blademaster / deep elf master archer / " ..
+         "deep elf annihilator w:" .. e .. " / deep elf demonologist w:" .. e
+end))
 
 mset(spec_fn(function ()
-               local d = 41 - you.depth()
-               return "base draconian w:" .. d .. " / nonbase draconian w:40"
-             end))
+  local d = math.max(5, 25 - you.zigs_completed())
+  local e = math.max(1, you.zigs_completed() + you.depth() - 11)
+  return "place:Vaults:$ w:" .. d * 2 .. " / place:Vaults:$ w:" .. d .. " / " ..
+         "glowing shapeshifter / sphinx w:5 / " ..
+         "titan w:" .. e .. " / golden dragon w:" .. e .. " / " ..
+         "ancient lich w:" .. e / 2 .. " / dread lich w:" .. e / 2
+end))
+
+mset(spec_fn(function ()
+  local d = 10 + you.zigs_completed() * 2
+  local e = 10 + you.zigs_completed() * 3
+  return "place:Crypt:$ 9 w:255 / ancient champion w:" .. d - 5 .. " / " ..
+         "curse skull w:" .. d - 5 .. " / profane servitor w:" .. d - 5 .. " / " ..
+         "bone dragon w:" .. d .. " / revenant w:" .. d .. " / " ..
+         "ancient lich w:" .. e / 2 .. " / dread lich w:" .. e / 2
+end))
+
+mset(spec_fn(function ()
+  local d = you.zigs_completed()
+  local e = 20 + you.depth() * 4 + you.zigs_completed() * 2
+  return "place:Tomb:$ 9 w:760 / bennu w:" .. d + 10 .. " / " ..
+         "mummy priest w:" .. d .. " / royal mummy w:" .. e
+end))
+
+mset(spec_fn(function ()
+  local d = 10 + you.zigs_completed() * 10
+  return "place:Abyss:$ w:1920 / corrupter w:" .. d * 2 .. " / " ..
+         "starcursed mass w:" .. d * 2 .. " / wretched star w:" .. d * 2 .. " / " ..
+         "bone dragon w:" .. d .. " / lich w:" .. d
+end))
+
+mset(with_props(spec_fn(function ()
+  local d = math.max(20, 455 - you.zigs_completed() * 3)
+  local e = 5 + you.zigs_completed() * 2
+  local f = 10 + you.zigs_completed() * 4
+  return "place:Coc:$ w:" .. d .. " / tormentor w:" .. e .. " / " ..
+         "shard shrike w:" .. e .. " / blizzard demon w:" .. f + 15 .. " / " ..
+         "ice fiend w:" .. f
+end), { weight = 5 }))
+
+mset(with_props(spec_fn(function ()
+  local d = math.max(20, 455 - you.zigs_completed() * 3)
+  local e = 5 + you.zigs_completed() * 2
+  local f = 10 + you.zigs_completed() * 4
+  return "place:Geh:$ w:" .. d .. " / hellion w:" .. e .. " / " ..
+         "hell knight w:" .. e .. " / balrug w:" .. f + 15 .. " / " ..
+         "brimstone fiend w:" .. f
+end), { weight = 5 }))
+
+mset(with_props(spec_fn(function ()
+  local d = math.max(20, 455 - you.zigs_completed() * 3)
+  local e = 5 + you.zigs_completed() * 2
+  local f = 10 + you.zigs_completed() * 4
+  return "place:Dis:$ w:" .. d .. " / iron dragon w:" .. e .. " / " ..
+         "iron giant w:" .. e .. " / war gargoyle w:" .. f + 15 .. " / " ..
+         "hell sentinel w:" .. f
+end), { weight = 5 }))
+
+mset(with_props(spec_fn(function ()
+  local d = math.max(20, 1840 - you.zigs_completed() * 12)
+  local e = 5 + you.zigs_completed() * 2
+  local f = 10 + you.zigs_completed() * 4
+  local g = 0 + you.zigs_completed()
+  return "place:Tar:$ w:" .. d .. " / shadow demon w:" .. e .. " / " ..
+         "curse toe w:" .. e .. " / reaper w:" .. f + 15 .. " / " ..
+         "tzitzimitl w:" .. f .. " / silent spectre w:" .. g
+end), { weight = 2 }))
+
+mset(with_props(spec_fn(function ()
+  local d = 10 + you.zigs_completed() * 2
+  local e = 10 + you.zigs_completed() * 3
+  return "fire elemental / hell hound / efreet / fire crab / " ..
+         "fire dragon w:" .. d .. " / fire giant w:" .. d .. " / " ..
+         "balrug w:" .. d .. " / orb of fire w:" .. e
+end), { weight = 2 }))
+
+mset(with_props(spec_fn(function ()
+  local d = 10 + you.zigs_completed() * 2
+  local e = 10 + you.zigs_completed() * 3
+  return "ice beast / ice devil / simulacrum / rime drake / " ..
+         "ice dragon w:" .. d .. " / frost giant w:" .. d .. " / " ..
+         "blizzard demon w:" .. d .. " / white draconian knight w:" .. e .. " / " ..
+         "shard shrike w:" .. e .. " / ice fiend w:" .. e
+end), { weight = 2 }))
+
+mset(with_props(spec_fn(function ()
+  local d = 10 + you.zigs_completed() * 2
+  local e = 10 + you.zigs_completed() * 3
+  return "raiju / air elemental / " ..
+         "shock serpent w:" .. d .. " / spark wasp w:" .. d .. " / " ..
+         "spriggan air mage w:" .. e .. " / storm dragon w:" .. e .. " / " ..
+         "titan w:" .. e .. " / electric golem w:" .. e
+end), { weight = 2 }))
+
+mset(with_props(spec_fn(function ()
+  local d = 20 + you.zigs_completed() * 2
+  local e = 20 + you.zigs_completed() * 3
+  local f = 20 + you.zigs_completed() * 4
+  return "gargoyle w:20 / earth elemental w:20 / boulder beetle w:20 / " ..
+         "torpor snail w:" .. d .. " / iron golem w:" .. d .. " / " ..
+         "war gargoyle w:" .. d .. " / stone giant w:" .. d .. " / " ..
+         "caustic shrike w:" .. d .. " / entropy weaver w:" .. d .. " / " ..
+         "crystal guardian w:" .. e .. " / iron dragon w:" .. e .. " / " ..
+         "iron giant w:" .. f .. " / hell sentinel w:" .. f
+end), { weight = 2 }))
+
+mset(with_props(spec_fn(function ()
+  local d = 20 + you.zigs_completed() * 2
+  local e = 20 + you.zigs_completed() * 3
+  return "ophan w:20 / apis w:20 / cherub w:20 / angel w:20 / " ..
+         "daeva w:" .. d ..  " / pearl dragon w:" .. e
+end), { weight = 2 }))
+
+mset(spec_fn(function ()
+  local d = math.max(2, math.floor((32 - you.depth()) / 5))
+  local e = math.min(8, math.floor((you.depth()) / 5) + 4)
+  local f = math.max(1, you.depth() + you.zigs_completed() - 5)
+  return "chaos spawn w:" .. d .. " / ugly thing w:" .. d .. " / " ..
+         "very ugly thing w:4 / apocalypse crab w:4 / " ..
+         "shapeshifter hd:16 w:" .. e .. " / glowing shapeshifter w:" .. e .. " / " ..
+         "killer klown w:8 / " ..
+         "greater demon w:2 / pandemonium lord w:" .. f
+end))
+
+mset(with_props(spec_fn(function ()
+  local d = math.max(2, math.floor((32 - you.depth()) / 5))
+  local e = math.min(8, math.floor((you.depth()) / 5) + 4)
+  local f = math.max(1, you.depth() + you.zigs_completed() - 5)
+  return "soul eater w:" .. d .. " / phantasmal warrior w:" .. d .. " / " ..
+         "deep elf death mage w:2 / shadow dragon w:4 / ghost crab w:4 / " ..
+         "eidolon w:" .. e .. " / revenant w:" .. e .. " / black sun w:4 / " ..
+         "curse skull w:4 / curse toe w:2 / player ghost w:" .. f
+end), { weight = 2 }))
+
+mset(with_props(spec_fn(function ()
+  local d = 20 + you.zigs_completed() * 3
+  local e = 20 + you.zigs_completed() * 4
+  local f = 20 + you.zigs_completed() * 5
+  return "cyclops w:20 / ettin w:20 / " ..
+         "stone giant w:" .. d .. " / " .. "fire giant w:" .. d .. " / " ..
+         "frost giant w:" .. d .. " / titan w:" .. f .. " / " ..
+         "juggernaut w:" .. e .. " / iron giant w:" .. e
+end), { weight = 2 }))
+
+mset(with_props(spec_fn(function ()
+  local d = 20 + you.zigs_completed() * 2
+  local e = 20 + you.zigs_completed() * 3
+  return "swamp drake w:20 / rime drake w:20 / wind drake / death drake w:20 / " ..
+         "wyvern / hydra / steam dragon w:20 / acid dragon w:20 / " ..
+         "swamp dragon w:" .. d .. " / fire dragon w:" .. d .. " / " ..
+         "ice dragon w:" .. d .. " / storm dragon w:" .. d .. " / " ..
+         "iron dragon w:" .. e .. " / quicksilver dragon w:" .. e .. " / " ..
+         "shadow dragon w:" .. e .. " / golden dragon w:" .. e
+end), { weight = 2 }))
+
+mset(spec_fn(function ()
+  local d = 41 - you.depth()
+  local e = 40 + you.zigs_completed()
+  return "base draconian w:" .. d .. " / nonbase draconian w:" .. e
+end))
+
+mset(with_props(spec_fn(function ()
+  local d = 20 + you.zigs_completed() * 2
+  local e = 20 + you.zigs_completed() * 3
+  return "centaur w:15 / yaktaur w:15 / cyclops w:15 / " ..
+         "centaur warrior w:" .. d .. " / faun w:" .. d .. " / " ..
+         "yaktaur captain w:" .. d .. " / satyr w:" .. d .. " / " ..
+         "stone giant w:" .. e .. " / naga sharpshooter w:" .. e .. " / " ..
+         "merfolk javelineer w:" .. e .. " / deep elf master archer w:" .. e
+end), { weight = 2 }))
+
+mset(with_props(spec_fn(function ()
+  local d = 10 + you.zigs_completed() * 2
+  local e = 10 + you.zigs_completed() * 4
+  return "wizard / necromancer / ogre mage w:5 / orc sorcerer w:5 / " ..
+         "naga mage / salamander mystic w:5 / merfolk aquamancer w:5 / " ..
+         "spriggan air mage w:" .. d - 5 .. " / nagaraja w:" .. d .. " / " ..
+         "deep elf annihilator w:" .. d .. " / deep elf sorcerer w:" .. d .. " / " ..
+         "tengu reaver w:" .. d .. " / draconian knight w:" .. d - 5 .. " / " ..
+         "draconian scorcher w:" .. d - 5 .. " / lich w:" .. d - 5 .. " / " ..
+         "ancient lich w:" .. d - 5 .. " / blood saint w:" .. d .. " / " ..
+         "draconian annihilator w:" .. e
+end), { weight = 2 }))
 
 local pan_lord_fn = zig_monster_fn("pandemonium lord")
-local pan_critter_fn = zig_monster_fn("place:Pan / greater demon / nonbase demonspawn w:4")
+local pan_critter_fn = zig_monster_fn("place:Pan w:" .. math.max(10, 100 - you.zigs_completed() * 4) ..
+                                      " / greater demon w:100 / nonbase demonspawn w:40")
 
 local function mons_panlord_gen(x, y, nth)
   if nth == 1 then
-    local d = math.max(1, you.depth() - 11)
     dgn.set_random_mon_list("place:Pan / greater demon / nonbase demonspawn w:4")
     return pan_lord_fn(x, y)
   else
@@ -327,14 +469,19 @@ local function mons_panlord_gen(x, y, nth)
   end
 end
 
-mset_if(depth_ge(8), mons_panlord_gen)
-mset_if(depth_ge(14), with_props("place:Snake:$ w:14 / place:Swamp:$ w:14 / " ..
-                      "place:Shoals:$ w:14 / place:Spider:$ w:14 / " ..
-                      "greater naga w:12 / guardian serpent w:8 / hydra w:5 / " ..
-                      "swamp dragon w:5 / tentacled monstrosity / " ..
-                      "merfolk aquamancer w:6 / merfolk javelineer w:8 / " ..
-                      "alligator snapping turtle w:6 / ghost moth w:8 / " ..
-                      "emperor scorpion w:8 / moth of wrath w:4", { weight = 5 }))
+mset(mons_panlord_gen)
+
+mset_if(depth_ge(14), with_props(spec_fn(function ()
+  local d = math.max(2, math.floor(14 - you.zigs_completed() / 2))
+  local e = 10 + you.zigs_completed()
+  return "place:Snake:$ w:" .. d .. " / place:Swamp:$ w:" .. d .. " / " ..
+         "place:Shoals:$ w:" .. d .. " / place:Spider:$ w:" .. d .. " / " ..
+         "nagaraja w:" .. e + 2 .. " / guardian serpent w:8 / " ..
+         "hydra w:5 / swamp dragon w:5 / tentacled monstrosity w:" .. e .. " / " ..
+         "merfolk aquamancer w:6 / merfolk javelineer w:" .. e - 2 .. " / " ..
+         "alligator snapping turtle w:6 / ghost moth w:" .. e - 2 .. " / " ..
+         "emperor scorpion w:8 / moth of wrath w:4"
+end), { weight = 5 }))
 
 function ziggurat_monster_creators()
   return util.map(monster_creator_fn, mons_populations)
@@ -375,8 +522,10 @@ end
 -- Function to find travel-safe squares, excluding closed doors.
 local dgn_passable = dgn.passable_excluding("closed_door")
 
-local function ziggurat_create_monsters(p, mfn)
-  local hd_pool = you.depth() * (you.depth() + 8) + 10
+local function ziggurat_create_monsters(entry, exit, mfn)
+  local depth = you.depth()
+  local completed = you.zigs_completed()
+  local hd_pool = math.floor(10 + (depth * (depth + 8 * (completed + 1) )) * (1 + completed * 0.085))
 
   local nth = 1
 
@@ -401,19 +550,23 @@ local function ziggurat_create_monsters(p, mfn)
   local function mons_place(point)
     if hd_pool <= 0 then
       return true
-    elseif not dgn.mons_at(point.x, point.y) then
+    elseif not dgn.mons_at(point.x, point.y) and entry ~= point then
       mons_do_place(point)
     end
   end
 
-  dgn.find_adjacent_point(p, mons_place, dgn_passable)
+  dgn.find_adjacent_point(exit, mons_place, dgn_passable)
 end
 
 local function ziggurat_create_loot_at(c)
-  -- Basically, loot grows linearly with depth.
+  -- Basically, loot grows linearly with depth & zigs finished.
   local depth = you.depth()
-  local nloot = depth
-  local nloot = depth + crawl.random2(math.floor(nloot * 0.5))
+  local completed = you.zigs_completed()
+  local nloot = depth + crawl.random2(math.floor(depth * 0.5))
+
+  if completed > 0 then
+    nloot = math.min(nloot + crawl.random2(math.floor(completed * depth / 4)), map_area() / 9)
+  end
 
   local function find_free_space(nspaces)
     local spaces = { }
@@ -430,13 +583,15 @@ local function ziggurat_create_loot_at(c)
   end
 
   -- dgn.good_scrolls is a list of items with total weight 1000
-  local good_loot = dgn.item_spec("* no_pickup no_mimic w:7000 / " .. dgn.good_scrolls)
-  local super_loot = dgn.item_spec("| no_pickup no_mimic w:7000 /" ..
-                                   "potion of experience no_pickup no_mimic w:200 /" ..
-                                   "potion of cure mutation no_pickup no_mimic w:200 /" ..
-                                   "potion of porridge no_pickup no_mimic w:100 /" ..
-                                   "wand of heal wounds no_pickup no_mimic w:10 / " ..
-                                   "wand of hasting no_pickup no_mimic w:10 / " ..
+  local good_loot = dgn.item_spec("* no_pickup w:7000 /" ..
+                                  dgn.good_scrolls)
+  local super_loot = dgn.item_spec("| no_pickup w:7000 /" ..
+                                   "potion of experience no_pickup w:190 q:1 /" ..
+                                   "potion of mutation no_pickup w:290 /" ..
+                                   "potion of cancellation q:5 no_pickup / " ..
+                                   "potion of heal wounds q:5 no_pickup / " ..
+                                   "potion of magic q:5 no_pickup / " ..
+                                   "potion of haste q:5 no_pickup / " ..
                                    dgn.good_scrolls)
 
   local loot_spots = find_free_space(nloot * 4)
@@ -470,83 +625,6 @@ local function ziggurat_create_loot_at(c)
   end
 end
 
--- Suitable for use in loot vaults.
-function ziggurat_loot_spot(e, key)
-  e.lua_marker(key, portal_desc { ziggurat_loot = "X" })
-  e.kfeat(key .. " = .")
-  e.marker("@ = lua:props_marker({ door_restrict=\"veto\" })")
-  e.kfeat("@ = +")
-end
-
-local has_loot_chamber = false
-
-local function ziggurat_create_loot_vault(entry, exit)
-  local inc = (exit - entry):sgn()
-
-  local connect_point = exit - inc * 3
-  local map = dgn.map_by_tag("ziggurat_loot_chamber")
-
-  if not map then
-    return exit
-  end
-
-  local function place_loot_chamber()
-    local res = dgn.place_map(map, true, true)
-    if res then
-      has_loot_chamber = true
-    end
-    return res
-  end
-
-  local function good_loot_bounds(map, px, py, xs, ys)
-    local vc = dgn.point(px + math.floor(xs / 2),
-                         py + math.floor(ys / 2))
-
-
-    local function safe_area()
-      local p = dgn.point(px, py)
-      local sz = dgn.point(xs, ys)
-      local floor = dgn.fnum("floor")
-      return dgn.rectangle_forall(p, p + sz - 1,
-                                  function (c)
-                                    return dgn.grid(c.x, c.y) == floor
-                                  end)
-    end
-
-    local linc = (exit - vc):sgn()
-    -- The map's positions should be at the same increment to the exit
-    -- as the exit is to the entrance, else reject the place.
-    return (inc == linc) and safe_area()
-  end
-
-  local function connect_loot_chamber()
-    return dgn.with_map_bounds_fn(good_loot_bounds, place_loot_chamber)
-  end
-
-  local res = dgn.with_map_anchors(connect_point.x, connect_point.y,
-                                   connect_loot_chamber)
-  if not res then
-    return exit
-  else
-    -- Find the square to drop the loot.
-    local lootx, looty = dgn.find_marker_position_by_prop("ziggurat_loot")
-
-    if lootx and looty then
-      return dgn.point(lootx, looty)
-    else
-      return exit
-    end
-  end
-end
-
-local function ziggurat_locate_loot(entrance, exit, jelly_protect)
-  if jelly_protect then
-    return ziggurat_create_loot_vault(entrance, exit)
-  else
-    return exit
-  end
-end
-
 local function ziggurat_place_pillars(c)
   local range = crawl.random_range
   local floor = dgn.fnum("floor")
@@ -617,6 +695,8 @@ local function ziggurat_stairs(entry, exit)
 
   if you.depth() < dgn.br_depth(you.branch()) then
     zigstair(exit.x, exit.y, "stone_stairs_down_i")
+  else
+    dgn.create_item(exit.x, exit.y, "figurine of a ziggurat")
   end
 
   zigstair(exit.x, exit.y + 1, "exit_ziggurat")
@@ -624,26 +704,19 @@ local function ziggurat_stairs(entry, exit)
 end
 
 local function ziggurat_furnish(centre, entry, exit)
-  has_loot_chamber = false
+  ziggurat_stairs(entry, exit)
+
   local monster_generation = choose_monster_set()
 
   if type(monster_generation.spec) == "string" then
     dgn.set_random_mon_list(monster_generation.spec)
   end
 
-  -- Identify where we're going to place loot, but don't actually put
-  -- anything down until we've placed pillars.
-  local lootspot = ziggurat_locate_loot(entry, exit,
-    monster_generation.jelly_protect)
+  ziggurat_place_pillars(centre)
 
-  if not has_loot_chamber then
-    -- Place pillars if we did not create a loot chamber.
-    ziggurat_place_pillars(centre)
-  end
+  ziggurat_create_loot_at(exit)
 
-  ziggurat_create_loot_at(lootspot)
-
-  ziggurat_create_monsters(exit, monster_generation.fn)
+  ziggurat_create_monsters(entry, exit, monster_generation.fn)
 
   local function needs_colour(p)
     return not dgn.in_vault(p.x, p.y)
@@ -695,7 +768,6 @@ local function ziggurat_rectangle_builder(e)
     entry, exit = exit, entry
   end
 
-  ziggurat_stairs(entry, exit)
   ziggurat_furnish(dgn.point(cx, cy), entry, exit)
 end
 
@@ -732,7 +804,6 @@ local function ziggurat_ellipse_builder(e)
     entry, exit = exit, entry
   end
 
-  ziggurat_stairs(entry, exit)
   ziggurat_furnish(dgn.point(cx, cy), entry, exit)
 end
 
@@ -778,7 +849,6 @@ local function ziggurat_hexagon_builder(e)
     entry, exit = exit, entry
   end
 
-  ziggurat_stairs(entry, exit)
   ziggurat_furnish(c, entry, exit)
 end
 
